# DoubleVIP release 4 by David Bowland
# ./addons/eventscripts/doublevip/doublevip.py

# >>> To configure this addon please see doublevip.cfg  (created when the addon is first loaded) <<<

"""
Picks one player on each team the team must protect to survive. Can also be used to create a VIP on only one team, with full support for as_ maps. Requires EZRestrict for weapon restrictions and ES 2.0.0.248+
"""


import cfglib
import es
import gamethread
import langlib
import os.path
import playerlib
import random

import psyco
psyco.full()


info = es.AddonInfo()
info.name     = 'DoubleVIP'
info.version  = '4'
info.basename = 'doublevip'
info.url      = 'http://addons.eventscripts.com/addons/view/' + info.basename
info.author   = 'SuperDave'


###


config = cfglib.AddonCFG(es.getAddonPath(info.basename) + '/' + info.basename + '.cfg')

config.text(info.name + ' release %s options' % info.version)
config.text('./addons/eventscripts/%(basename)s/%(basename)s.cfg' % {'basename':info.basename})
config.text(info.url)
config.text('')
config.text('Load this addon with: es_load ' + info.basename)
config.text('\n')
config.text('To configure language settings for %s please see %s_languages.ini' % (info.name, info.basename))
config.text('\n')
config.text('This addon is compatible with VIP maps (maps designed for VIP Mod).')
config.text('\n')
config.text('This addon requires EZRestrict loaded if any weapons are restricted. EZRestrict can be found here:')
config.text('http://addons.eventscripts.com/addons/view/ezrestrict')
config.text('\n')

config.text('***** General VIP options *****')
cvar_vip_health = config.cvar('doublevip_vip_health', 250, 'Starting health of VIPs')
cvar_bot_avoid  = config.cvar('doublevip_bot_avoid',  1,   '0 = all players have an equal chance of becoming VIP, 1 = human players will be preferred for VIP')
cvar_bot_speed  = config.cvar('doublevip_bot_speed',  4,   'Speed of VIP bots from 0 (half speed) to 5 (full speed)')
config.text('\n')

config.text('***** VIP weapon options *****')
cvar_vip_restrict_primary   = config.cvar('doublevip_vip_restrict_primary',   1, '0 = VIPs are allowed primary weapons, 1 = VIPs will be restricted from primary weapons')
cvar_vip_restrict_secondary = config.cvar('doublevip_vip_restrict_secondary', 0, '0 = VIPs are allowed secondary weapons, 1 = VIPs will be restricted from secondary weapons')
cvar_vip_restrict_grenades  = config.cvar('doublevip_vip_restrict_grenades',  0, '0 = VIPs are allowed grenades, 1 = VIPs will be restricted from grenades')
cvar_vip_restrict_remove    = config.cvar('doublevip_vip_restrict_remove',    0, '0 = VIPs will drop restricted weapons, 1 = restrcited weapons picked up by a VIP will be removed')
config.text('\n')

config.text('***** VIP color options *****')
config.text('Normal color can be achieved by setting all color options to 255.')
cvar_vip_transparency = config.cvar('doublevip_vip_transparency', 75, 'Transparency of VIPs from 0 (invisible) to 255 (opaque/no change)')
config.cvar('doublevip_vip_color_t_red',    200, 'Intensity of red for terrorist VIP (0-255)')
config.cvar('doublevip_vip_color_t_green',  0,   'Intensity of green for terrorist VIP (0-255)')
config.cvar('doublevip_vip_color_t_blue',   0,   'Intensity of blue for terrorist VIP (0-255)')
config.cvar('doublevip_vip_color_ct_red',   0,   'Intensity of red for counter-terrorist VIP (0-255)')
config.cvar('doublevip_vip_color_ct_green', 0,   'Intensity of green for counter-terrorist VIP (0-255)')
config.cvar('doublevip_vip_color_ct_blue',  200, 'Intensity of blue for counter-terrorist VIP (0-255)')
config.text('\n')

config.text('***** Miscellaneous option *****')
cvar_removeobjectives = config.cvar('doublevip_removeobjectives', 1,   '0 = no change, 1 = map objectives will be removed (does not remove VIP escape zones)')
cvar_escapecheck      = config.cvar('doublevip_escapecheck',      1.5, 'Number of seconds between checking for escaped VIPs (assassination maps only)--set to 0 to eliminate')
config.text('Note: DoubleVIP must be loaded when the map starts to recognize escape zones.')
cvar_singlevip        = config.cvar('doublevip_singlevip',        0,   '0 = one VIP per team, 1 = only terrorists have a VIP, 2 = only counter-terrorists have a VIP')
config.text('\n')

config.text('***** Announcement options *****')
cvar_announce_hud     = config.cvar('doublevip_announce_hud',     1, '0 = no change, 1 = announce addon objectives to players at round start with a hud hint')
cvar_announce_chat    = config.cvar('doublevip_announce_chat',    0, '0 = no change, 1 = announce to players when they are picked to be a VIP in chat area, 2 = announce to team members when a player is picked to be a VIP in chat area, 3 = announce to all players when a player is picked to be a VIP in chat area')
cvar_announce_center  = config.cvar('doublevip_announce_center',  1, '0 = no change, 1 = announce to players when they are picked to be a VIP in center text, 2 = announce to team members when a player is picked to be a VIP in center text')
cvar_announce_console = config.cvar('doublevip_announce_console', 2, '0 = no change, 1 = announce to player when they are picked to be a VIP on the console, 2 = announce to team members when a player is picked to be a VIP on the console, 3 = announce to all players when a player is picked to be a VIP on the console')
config.text('\n')

config.text('***** Event documentation *****')
config.text('')
config.text('The following events will fire when DoubleVIP is loaded:')
config.text('')
config.text('vip_escaped')
config.text('- Default Valve event, only fires for first VIP to escape')
config.text('   event_var(userid)          - VIP who escaped')
config.text('')
config.text('vip_picked')
config.text('   event_var(userid)          - Player picked to be a VIP')
config.text('   event_var(team)            - Team number of VIP')
config.text('')
config.text('vip_killed')
config.text('- Default Valve event')
config.text('   event_var(userid)          - VIP who was killed')
config.text('   event_var(attacker)        - Player who killed the VIP')
config.text('')
config.text('These events also provide event_var(es_steamid), event_var(es_username), and all the other event_vars that accompany')
config.text('event_var(userid). More information can be found here: http://www.eventscripts.com/pages/Extended_event_variables')

config.write()


###


class DoubleVIP(object):
   events = {'vip_picked':{'userid':'short', 'team':'short'}}

   def __init__(self):
      self.vips      = {}
      self.escapes   = set()
      self.teleports = []

   @staticmethod
   def clearDelays():
      gamethread.cancelDelayed('doublevip_drop')
      gamethread.cancelDelayed('doublevip_escape')

   def registerClientFilter(self):
      es.addons.registerClientCommandFilter(self._cc_filter)

   def unregisterClientFilter(self):
      es.addons.unregisterClientCommandFilter(self._cc_filter)

   def loadEvents(self):
      if os.path.isfile(es.getAddonPath('doublevip') + '/doublevip.res'):
         es.loadevents('declare', 'addons/eventscripts/doublevip/doublevip.res')
      else:
         if not os.path.isfile(es.getAddonPath('doublevip') + 'es_doublevip_events_db.txt'):
            es.keygroupcreate('doublevip_events')
            for event_name in self.events:
               es.keycreate('doublevip_events', event_name)
               for event_var in self.events[event_name]:
                  es.keysetvalue('doublevip_events', event_name, event_var, self.events[event_name][event_var])
            es.keygroupsave('doublevip_events', '|doublevip')
            es.keygroupdelete('doublevip_events')
         es.loadevents('declare', 'addons/eventscripts/doublevip/es_doublevip_events_db.txt')
         es.dbgmsg(0, 'DoubleVIP: Unable to load doublevip.res! Please ensure it is in the ./doublevip/ directory.')


   def roundStart(self):
      userid = es.getuserid()
      if not userid: return

      if int(cvar_removeobjectives):
         es.server.queuecmd('es_xfire %s weapon_c4 kill' % userid)
         es.server.queuecmd('es_xfire %s func_bomb_target kill' % userid)
         es.server.queuecmd('es_xfire %s hostage_entity kill' % userid)
         es.server.queuecmd('es_xfire %s func_hostage_rescue kill' % userid)

      vip_restrictions = []
      if int(cvar_vip_restrict_primary):
         vip_restrictions.append('#primary')
      if int(cvar_vip_restrict_secondary):
         vip_restrictions.append('#secondary')
      if int(cvar_vip_restrict_grenades):
         vip_restrictions.append('#grenade')

      bot_speed  = int(cvar_bot_speed)
      vip_alpha  = int(cvar_vip_transparency)
      vip_health = int(cvar_vip_health)
      avoid_bots = int(cvar_bot_avoid)
      lastvips   = self.vips.keys()
      playerlist = es.getUseridList()
      entitylist = es.createentitylist()

      singlevip = int(cvar_singlevip)
      vipteams  = []
      if singlevip <> 2:
         vipteams.append((2, 't'))
      if singlevip <> 1:
         vipteams.append((3, 'ct'))

      chat    = int(cvar_announce_chat)
      center  = int(cvar_announce_center)
      console = int(cvar_announce_console)
      hud     = int(cvar_announce_hud)

      # Pick new VIPs
      self.vips.clear()
      for team, teamtag in vipteams:
         choices = filter(lambda x: es.getplayerteam(x) == team, playerlist)
         if not choices: continue

         # Ignore recent VIPs
         newchoices = filter(lambda x: x not in lastvips, choices)
         if newchoices:
            choices = newchoices

         # Ignore bots if necessary
         if avoid_bots:
            newchoices = filter(lambda x: es.getplayersteamid(x) <> 'BOT', choices)
            if newchoices:
               choices = newchoices

         # Select VIP
         userid = random.choice(choices)
         self.restrict(userid, vip_restrictions)
         currentvip = self.vips[userid] = {'restrictions':vip_restrictions, 'team':team, 'alpha':vip_alpha}
         es.setplayerprop(userid, 'CBasePlayer.m_iHealth', vip_health)
         if es.getplayersteamid(userid) == 'BOT' and bot_speed <> 5:
            es.setplayerprop(userid, 'CBasePlayer.localdata.m_flLaggedMovementValue', (5 + bot_speed) / 10.)

         color = currentvip['color'] = tuple(int(es.ServerVar('doublevip_vip_color_%s_%s' % (teamtag, x))) for x in ('red', 'green', 'blue'))
         es.server.queuecmd('es_xfire %s !self addoutput "rendermode 1"' % userid)
         if vip_alpha <> 255:
            es.server.queuecmd('es_xfire %s !self alpha %s' % (userid, vip_alpha))
         if color <> (255, 255, 255):
            es.server.queuecmd('es_xfire %s !self color "%s %s %s"' % ((userid,) + color))

         handle  = es.getplayerhandle(userid)
         weapons = currentvip['weapons'] = set()
         for index in entitylist:
            if entitylist[index]['classname'].startswith('weapon_'):
               if self.getOwner(index) == handle:
                  weapons.add(index)
                  self.colorEntity(index, color, vip_alpha)

         es.event('initialize', 'vip_picked')
         es.event('setint',     'vip_picked', 'userid', userid)
         es.event('setint',     'vip_picked', 'team',   team)
         es.event('fire',       'vip_picked')

         # Announce VIP
         vipname  = {'vip':es.getplayername(userid)}
         textname = 'vip ' + teamtag
         for loop_userid in playerlist:
            player_lang = playerlib.getPlayer(userid).get('lang')
            loop_team   = team == 5 - es.getplayerteam(loop_userid)

            if chat >= 1 and loop_userid == userid:
               es.tell(userid, '#multi', lang_text('vip player', lang=player_lang))
            elif chat >= 2 and not loop_team:
               es.tell(loop_userid, '#multi', lang_text(textname, vipname, player_lang))
            elif chat >= 3:
               es.tell(loop_userid, '#multi', lang_text(textname, vipname, player_lang))

            if center >= 1 and loop_userid == userid:
               es.centertell(userid, remove_tags(lang_text('vip player', lang=player_lang)))
            elif center >= 2 and not loop_team:
               es.centertell(loop_userid, remove_tags(lang_text(textname, vipname, player_lang)))

            if console >= 1 and loop_userid == userid:
               es.cexec(loop_userid, 'echo ' + remove_tags(lang_text('vip player', lang=player_lang)))
            elif console >= 2 and not loop_team:
               es.cexec(loop_userid, 'echo ' + remove_tags(lang_text(textname, vipname, player_lang)))
            elif console >= 3:
               es.cexec(loop_userid, 'echo ' + remove_tags(lang_text(textname, vipname, player_lang)))

      if int(cvar_announce_hud):
         for userid in playerlist:
            team = es.getplayerteam(userid)
            if team > 1:
               player_lang = playerlib.getPlayer(userid).get('lang')
               if userid in self.vips:
                  self.hudHint(userid, lang_text('objective vip', lang=player_lang))
               elif (singlevip == 1 and team == 3) or (singlevip == 2 and team == 2):
                  self.hudHint(userid, lang_text('eliminate only objective', lang=player_lang))
               elif (singlevip == 1 and team == 2) or (singlevip == 2 and team == 3):
                  self.hudHint(userid, lang_text('protect only objective', lang=player_lang))
               else:
                  self.hudHint(userid, lang_text('objective double', lang=player_lang))

      delay = float(cvar_escapecheck)
      gamethread.cancelDelayed('doublevip_escape')
      if delay and self.escapes:
         gamethread.delayedname(delay, 'doublevip_escape', self._checkEscape)

   def roundEnd(self):
      entitylist = es.createentitylist()
      for userid in self.vips:
         vip = self.vips[userid]
         self.unrestrict(userid, vip['restrictions'])

         color, alpha = vip['color'], vip['alpha']
         if color <> (255, 255, 255):
            es.server.queuecmd('es_xfire %s !self color "255 255 255"' % userid)
         if alpha <> 255:
            es.server.queuecmd('es_xfire %s !self alpha 255' % userid)

         for index in vip['weapons']:
            if index in entitylist:
               self.colorEntity(index)

   def itemPickup(self, userid, item):
      if userid not in self.vips: return

      vip     = self.vips[userid]
      weapons = vip['weapons']
      handle  = es.getplayerhandle(userid)
      for index in es.createentitylist('weapon_' + item):
         if self.getOwner(index) == handle:
            weapons.add(index)
            self.colorEntity(index, vip['color'], vip['alpha'])
            break

   def restrictedPickup(self, userid, index):
      if userid not in self.vips or not int(cvar_vip_restrict_remove): return

      try:
         es.import_addon('ezrestrict').saferemove(index)
      except KeyError: pass

   def playerDeath(self, userid, attacker=0):
      if userid not in self.vips: return
      self.slayTeam(self.vips[userid]['team'], attacker if attacker else userid)

      es.event('initialize', 'vip_killed')
      es.event('setint',     'vip_killed', 'userid',   userid)
      es.event('setint',     'vip_killed', 'attacker', attacker)
      es.event('fire',       'vip_killed')


   def restrict(self, userid, restrictions):
      if not restrictions: return

      try:
         es.import_addon('ezrestrict').restrict(userid, restrictions)
      except KeyError:
         es.dbgmsg(0, 'DoubleVIP: Cannot enforce weapon restrictions because EZRestrict is not loaded.')

   def unrestrict(self, userid, restrictions):
      if not restrictions or not es.exists('userid', userid): return

      try:
         es.import_addon('ezrestrict').unrestrict(userid, restrictions)
      except KeyError: pass


   def _cc_filter(self, userid, args):
      if args[0].lower() == 'drop' and userid in self.vips:
         handle = es.getplayerhandle(userid)
         for index in es.createentitylist(es.createplayerlist(userid)[userid]['weapon']):
            if self.getOwner(index) == handle:
               gamethread.delayedname(0.1, 'doublevip_drop', self._drop, (userid, index))
               break
      return True

   def _drop(self, userid, index):
      if not es.createentitylist(index): return

      if self.getOwner(index) == -1:
         self.colorEntity(index)
         weapons = self.vips[userid]['weapons']
         if index in weapons:
            weapons.remove(index)


   def resetEscapeZones(self):
      self.escapes.clear()
      self.teleports[:] = []

   def addEscapeZone(self, mins, maxs):
      self.escapes.add((mins, maxs))

   def addTeleport(self, location):
      if location not in self.teleports:
         self.teleports.append(location)

   def _checkEscape(self):
      for userid in self.vips:
         pos = es.getplayerlocation(userid)
         for mins, maxs in self.escapes:
            es.msg('min %s max %s pos %s' % (mins, maxs, pos))
            if len(filter(lambda x: mins[x] <= pos[x] <= maxs[x], range(3))) == 3:
               self.slayTeam(5 - self.vips[userid]['team'], userid)

               if self.teleports:
                  es.server.queuecmd('es_xsetpos %s %s' % (userid, random.choice(self.teleports)))

               es.event('initialize', 'vip_escaped')
               es.event('setint',     'vip_escaped', 'userid', userid)
               es.event('fire',       'vip_escaped')
               return # Only one VIP can escape, no more looping

      delay = float(cvar_escapecheck)
      if delay:
         gamethread.delayedname(delay, 'doublevip_escape', self._checkEscape)


   @staticmethod
   def slayTeam(team, attacker):
      for userid in filter(lambda x: es.getplayerteam(x) == team and not es.getplayerprop(x, 'CBasePlayer.pl.deadflag'), es.getUseridList()):
         es.server.queuecmd('es_xfire %s !self addoutput \"targetname doublevipslay\"' % userid)

         es.server.queuecmd('es_xfire %s point_hurt kill' % attacker)
         es.server.queuecmd('es_xentcreate %s point_hurt' % attacker)
         es.server.queuecmd('es_xfire %s point_hurt addoutput \"damagetype 32\"' % attacker)
         es.server.queuecmd('es_xfire %s point_hurt addoutput \"damagetarget doublevipslay\"' % attacker)
         es.server.queuecmd('es_xfire %s point_hurt addoutput \"damage %s\"' % (attacker, es.getplayerprop(userid, 'CBasePlayer.m_iHealth')))
         es.server.queuecmd('es_xfire %s point_hurt hurt' % attacker)

         es.server.queuecmd('es_xfire %s !self addoutput \"targetname none\"' % userid)

   @staticmethod
   def getOwner(index):
      return es.getindexprop(index, 'CBaseEntity.m_hOwnerEntity')

   @staticmethod
   def colorEntity(index, color=(255, 255, 255), alpha=255):
      es.setindexprop(index, 'CBaseEntity.m_nRenderMode', es.getindexprop(index, 'CBaseEntity.m_nRenderMode') | 1)
      es.setindexprop(index, 'CBaseEntity.m_nRenderFX', es.getindexprop(index, 'CBaseEntity.m_nRenderFX') | 256)
      es.setindexprop(index, 'CBaseEntity.m_clrRender', color[0] + (color[1] << 8) + (color[2] << 16) + (alpha << 24))

   @staticmethod
   def hudHint(userid, message):
      es.usermsg('create', 'doublevip_hud', 'HintText')
      es.usermsg('write', 'short', 'doublevip_hud', -1)
      es.usermsg('write', 'string', 'doublevip_hud', message)
      es.usermsg('send', 'doublevip_hud', userid)
      es.usermsg('delete', 'doublevip_hud')

vips = DoubleVIP()


def escapezone_cmd():
   """
   vip_escapezone <minX> <minY> <minZ> <maxX> <maxY> <maxZ>
   """
   if es.getargc() == 7:
      vips.addEscapeZone(tuple(map(lambda x: float(es.getargv(x)), range(1, 4))), tuple(map(lambda x: float(es.getargv(x)), range(4, 7))))

   else:
      es.dbgmsg(0, 'Syntax: %s <minX> <minY> <minZ> <maxX> <maxY> <maxZ>' % es.getargv(0).lower())

if not es.exists('command', 'vip_escapezone'):
   es.regcmd('vip_escapezone', 'doublevip/escapezone_cmd', 'vip_escapezone <minX> <minY> <minZ> <maxX> <maxY> <maxZ>\nServer command used by assassination maps to designate escape zone')
if not es.exists('command', 'vip_escapezone2'):
   es.regcmd('vip_escapezone2', 'doublevip/escapezone_cmd', 'vip_escapezone2 <minX> <minY> <minZ> <maxX> <maxY> <maxZ>\nServer command used by assassination maps to designate escape zone')


def teleport_cmd():
   """
   vip_teleport <x> <y> <z>
   """
   if es.getargc() == 4:
      vips.addTeleport(' '.join(map(lambda x: float(es.getargv(x)), range(1, 4))))

   else:
      es.dbgmsg(0, 'Syntax: vip_teleport <x> <y> <z>')

if not es.exists('command', 'vip_teleport'):
   es.regcmd('vip_teleport', 'doublevip/teleport_cmd', 'vip_teleport <x> <y> <z>\nServer command used by assassination maps to designate VIP teleport locations')


###


def getLangText():
   lang_path = es.getAddonPath(info.basename) + '/%s_languages.ini' % info.basename
   if os.path.isfile(lang_path):
      return langlib.Strings(lang_path)

   else:
      es.dbgmsg(0, '%(name)s: Unable to load %(basename)s_languages.ini! Please ensure it is in the ./%(basename)s/ directory.' % {'name':info.name, 'basename':info.basename})
      return lambda x, y, z: '%s: %s_languages.ini not found' % (info.name, info.basename)

lang_text = getLangText()


def remove_tags(text):
   """Removes #lightgreen, #green and #default tags from the supplied string"""
   return text.replace('#lightgreen', '').replace('#green', '').replace('#default', '')


###


def load():
   config.execute()

   vips.loadEvents()
   vips.registerClientFilter()


def es_map_start(event_var):
   vips.loadEvents()
   vips.clearDelays()
   vips.resetEscapeZones()


def round_start(event_var):
   vips.roundStart()


def round_end(event_var):
   vips.roundEnd()


def item_pickup(event_var):
   vips.itemPickup(int(event_var['userid']), event_var['item'])


def player_death(event_var):
   vips.playerDeath(int(event_var['userid']), int(event_var['attacker']))


def player_disconnect(event_var):
   vips.playerDeath(int(event_var['userid']))


def ezrestricted_pickup(event_var):
   vips.restrictedPickup(int(event_var['userid']), int(event_var['weaponindex']))


def unload():
   vips.roundEnd()
   vips.clearDelays()
   vips.unregisterClientFilter()

   for cvar in config.getCvars():
      es.ServerVar(cvar).set(0)